from typing import cast, Any, Optional
import jsii
import pytest
import re

from jsii.errors import JSIIError
import jsii_calc
from jsii_calc.module2702 import IVpc, Vpc, IBaz, Baz
from jsii_calc.jsii3656 import OverrideMe


class TestErrorHandling:
    def test_jsii_error(self):
        obj = jsii_calc.Calculator()

        with pytest.raises(
            JSIIError, match="Class jsii-calc.Calculator doesn't have a method"
        ):
            jsii.kernel.invoke(obj, "nonexistentMethod")

    def test_inheritance_maintained(self):
        """Check that for JSII struct types we can get the inheritance tree in some way."""
        # inspect.getmro() won't work because of TypedDict, but we add another annotation
        bases = find_struct_bases(jsii_calc.DerivedStruct)

        base_names = [b.__name__ for b in bases]

        assert base_names == ["DerivedStruct", "MyFirstStruct"]

    def test_descriptive_error_when_passing_function(self):
        obj = jsii_calc.Calculator()

        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument value must be one of (int, float); got method instead"
            ),
        ):
            obj.add(cast(Any, self.test_descriptive_error_when_passing_function))

    def test_implements_interface(self) -> None:
        """Checks that jsii-generated classes correctly implement the relevant jsii-generated interfaces."""

        def vpc_interface_func(v: IVpc) -> None:
            assert v is not None

        vpc = Vpc()
        vpc_interface_func(vpc)

        def baz_interface_func(b: IBaz) -> None:
            assert b is not None

        baz = Baz()
        baz_interface_func(baz)


def test_overrides_method_with_kwargs() -> None:
    class Overridden(OverrideMe):
        def implement_me(
            self, *, name: str, count: Optional[jsii.Number] = None
        ) -> bool:
            return name == "John Doe" and count is None

    assert OverrideMe.call_abstract(Overridden())


def find_struct_bases(x):
    ret = []
    seen = set([])

    def recurse(s):
        if s not in seen:
            ret.append(s)
            seen.add(s)
            bases = getattr(s, "__jsii_struct_bases__", [])
            for base in bases:
                recurse(base)

    recurse(x)
    return ret



import json

from ._compat import importlib_resources

# Load our version number and other metadata.
_meta = json.loads(
    importlib_resources.files("jsii").joinpath("_metadata.json").read_text()
)

__version__ = _meta["version"]
__jsii_runtime_version__ = _meta["jsii-runtime"]["version"]


__all__ = ["__version__", "__jsii_runtime_version__"]



import platform

from datetime import datetime, timezone
from typing import cast, List

import pytest

import jsii

from json import loads

from jsii_calc import (
    AbstractClassReturner,
    AbstractSuite,
    Add,
    AllTypes,
    AllTypesEnum,
    AmbiguousParameters,
    AsyncVirtualMethods,
    Bell,
    Calculator,
    ClassWithPrivateConstructorAndAutomaticProperties,
    ConfusingToJackson,
    ConsumerCanRingBell,
    ConsumePureInterface,
    ConstructorPassesThisOut,
    DataRenderer,
    Demonstrate982,
    DisappointingCollectionSource,
    DoNotOverridePrivates,
    DoubleTrouble,
    Entropy,
    GreetingAugmenter,
    IBellRinger,
    IConcreteBellRinger,
    IFriendlier,
    IFriendlyRandomGenerator,
    IRandomNumberGenerator,
    InterfaceCollections,
    IInterfaceWithProperties,
    IStructReturningDelegate,
    IWallClock,
    Isomorphism,
    JsiiAgent,
    JSObjectLiteralForInterface,
    JSObjectLiteralToNative,
    JsonFormatter,
    Multiply,
    Negate,
    NodeStandardLibrary,
    NullShouldBeTreatedAsUndefined,
    NumberGenerator,
    ObjectWithPropertyProvider,
    PartiallyInitializedThisConsumer,
    Polymorphism,
    Power,
    PythonReservedWords,
    ReferenceEnumFromScopedPackage,
    ReturnsPrivateImplementationOfInterface,
    Statics,
    Sum,
    SyncVirtualMethods,
    UsesInterfaceWithProperties,
    composition,
    EraseUndefinedHashValues,
    EraseUndefinedHashValuesOptions,
    VariadicMethod,
    RootStructValidator,
    StructPassing,
    TopLevelStruct,
    SecondLevelStruct,
    StructA,
    StructB,
    StructUnionConsumer,
    SomeTypeJsii976,
    StructParameterType,
    AnonymousImplementationProvider,
    UpcasingReflectable,
    PromiseNothing,
)
from jsii_calc.cdk16625 import Cdk16625
from jsii_calc.cdk22369 import AcceptsPath
from jsii_calc.python_self import (
    ClassWithSelf,
    ClassWithSelfKwarg,
)
from jsii_calc.submodule.isolated import Kwargs
from jsii_calc.submodule.child import SomeEnum
from scope.jsii_calc_lib import IFriendly, EnumFromScopedModule, Number
from scope.jsii_calc_lib.custom_submodule_name import IReflectable, ReflectableEntry
from scope.jsii_calc_lib.deprecation_removal import InterfaceFactory

# Note: The names of these test functions have been chosen to map as closely to the
#       Java Compliance tests as possible.
# Note: While we could write more expressive and better tests using the functionality
#       provided to us by pytest, we are making these tests match the Java Compliance
#       Tests as closely as possible to make keeping them in sync easier.


class DerivedFromAllTypes(AllTypes):
    pass


class OverrideAsyncMethods(AsyncVirtualMethods):
    def override_me(self, mult):
        return self.foo() * 2

    def foo(self) -> int:
        """
        Implement another method, which doesn't override anything in the base class.
        This should obviously be possible.
        """
        return 2222


class OverrideAsyncMethodsByBaseClass(OverrideAsyncMethods):
    pass


class OverrideCallsSuper(AsyncVirtualMethods):
    def override_me(self, mult):
        super_ret = super().override_me(mult)
        return super_ret * 10 + 1


class TwoOverrides(AsyncVirtualMethods):
    def override_me(self, mult):
        return 666

    def override_me_too(self):
        return 10


class SyncOverrides(SyncVirtualMethods):
    multiplier = 1
    return_super = False
    call_async = False
    another_the_property = None

    def virtual_method(self, n):
        if self.return_super:
            return super().virtual_method(n)

        if self.call_async:
            obj = OverrideAsyncMethods()
            return obj.call_me()

        return 5 * n * self.multiplier

    @property
    def the_property(self):
        return "I am an override!"

    @the_property.setter
    def the_property(self, value):
        self.another_the_property = value


@jsii.implements(IFriendly)
@jsii.implements(IRandomNumberGenerator)
class SubclassNativeFriendlyRandom(Number):
    def __init__(self):
        super().__init__(908)
        self.next_number = 100

    def hello(self):
        return "SubclassNativeFriendlyRandom"

    def next(self):
        next_ = self.next_number
        self.next_number += 100
        return next_


@jsii.implements(IFriendlyRandomGenerator)
class PureNativeFriendlyRandom:
    """
    In this case, the class does not derive from the JsiiObject hierarchy. It means
    that when we pass it along to javascript, we won't have an objref. This should
    result in creating a new empty javascript object and applying the overrides.

    The newly created objref will need to be stored somewhere (in the engine's object
    map) so that subsequent calls won't create a new object every time.
    """

    next_number = 1000

    def next(self):
        n = self.next_number
        self.next_number += 1000
        return n

    def hello(self):
        return "I am a native!"


class AddTen(Add):
    def __init__(self, value):
        super().__init__(Number(value), Number(10))


class MulTen(Multiply):
    def __init__(self, value):
        super().__init__(Number(value), Number(10))


def test_primitiveTypes():
    types = AllTypes()

    # boolean
    types.boolean_property = True
    assert types.boolean_property

    # string
    types.string_property = "foo"
    assert types.string_property == "foo"

    # number
    types.number_property = 1234
    assert types.number_property == 1234

    # date
    types.date_property = datetime.fromtimestamp(123 / 1000.0, tz=timezone.utc)
    assert types.date_property == datetime.fromtimestamp(123 / 1000.0, tz=timezone.utc)

    # json
    types.json_property = {"Foo": {"bar": 123}}
    assert types.json_property.get("Foo") == {"bar": 123}


def test_dates():
    types = AllTypes()

    # strong type
    types.date_property = datetime.fromtimestamp(123 / 1000.0, tz=timezone.utc)
    assert types.date_property == datetime.fromtimestamp(123 / 1000.0, tz=timezone.utc)

    # weak type
    types.any_property = datetime.fromtimestamp(999 / 1000.0, tz=timezone.utc)
    assert types.any_property == datetime.fromtimestamp(999 / 1000.0, tz=timezone.utc)


def test_collectionTypes():
    types = AllTypes()

    # array
    types.array_property = ["Hello", "World"]
    assert types.array_property[1] == "World"

    # map
    map_ = {}
    map_["Foo"] = Number(123)
    types.map_property = map_
    # TODO: No Assertion?


def test_dynamicTypes():
    types = AllTypes()

    # boolean
    types.any_property = False
    assert not types.any_property

    # string
    types.any_property = "String"
    assert types.any_property == "String"

    # number
    types.any_property = 12
    assert types.any_property == 12

    # date
    types.any_property = datetime.fromtimestamp(1234 / 1000.0, tz=timezone.utc)
    assert types.any_property == datetime.fromtimestamp(1234 / 1000.0, tz=timezone.utc)

    # json (notice that when deserialized, it is deserialized as a map).
    types.any_property = {"Goo": ["Hello", {"World": 123}]}
    got = types.any_property.get("Goo")
    assert got is not None
    assert got[1].get("World") == 123

    # array
    types.any_property = ["Hello", "World"]
    assert types.any_property[0] == "Hello"
    assert types.any_property[1] == "World"

    # array of any
    types.any_array_property = ["Hybrid", Number(12), 123, False]
    assert types.any_array_property[2] == 123

    # map
    map_ = {}
    map_["MapKey"] = "MapValue"
    types.any_property = map_
    assert types.any_property.get("MapKey") == "MapValue"

    # map of any
    map_["Goo"] = 19_289_812
    types.any_map_property = map_
    assert types.any_map_property.get("Goo") == 19_289_812

    # classes
    mult = Multiply(Number(10), Number(20))
    types.any_property = mult
    assert types.any_property is mult
    assert isinstance(types.any_property, Multiply)
    assert types.any_property.value == 200


def test_unionTypes():
    types = AllTypes()

    # single valued property
    types.union_property = 1234
    assert types.union_property == 1234

    types.union_property = "Hello"
    assert types.union_property == "Hello"

    types.union_property = Multiply(Number(2), Number(12))
    assert types.union_property.value == 24

    # map
    map_ = {}
    map_["Foo"] = Number(99)
    types.union_map_property = map_
    # TODO: No Assertion?

    # array
    types.union_array_property = [123, Number(33)]
    assert cast(Number, types.union_array_property[1]).value == 33


def test_createObjectAndCtorOverloads():
    Calculator()
    Calculator(maximum_value=10)


def test_getSetPrimitiveProperties():
    number = Number(20)

    assert number.value == 20
    assert number.double_value == 40
    assert Negate(Add(Number(20), Number(10))).value == -30
    assert Multiply(Add(Number(5), Number(5)), Number(2)).value == 20
    assert Power(Number(3), Number(4)).value == 3**4
    assert Power(Number(999), Number(1)).value == 999
    assert Power(Number(999), Number(0)).value == 1


def test_callMethods():
    calc = Calculator()

    calc.add(10)
    assert calc.value == 10

    calc.mul(2)
    assert calc.value == 20

    calc.pow(5)
    assert calc.value == 20**5

    calc.neg()
    assert calc.value == -3_200_000


def test_unmarshallIntoAbstractType():
    calc = Calculator()
    calc.add(120)

    assert calc.curr.value == 120


def test_getAndSetNonPrimitiveProperties():
    calc = Calculator()
    calc.add(3_200_000)
    calc.neg()
    calc.curr = Multiply(Number(2), calc.curr)

    assert calc.value == -6_400_000


def test_getAndSetEnumValues():
    calc = Calculator()
    calc.add(9)
    calc.pow(3)

    CompositeOperation = composition.CompositeOperation

    assert calc.string_style == CompositeOperation.CompositionStringStyle.NORMAL

    calc.string_style = CompositeOperation.CompositionStringStyle.DECORATED

    assert calc.string_style == CompositeOperation.CompositionStringStyle.DECORATED
    assert calc.to_string() == "<<[[{{(((1 * (0 + 9)) * (0 + 9)) * (0 + 9))}}]]>>"


def test_useEnumFromScopedModule():
    obj = ReferenceEnumFromScopedPackage()
    assert obj.foo == EnumFromScopedModule.VALUE2
    obj.foo = EnumFromScopedModule.VALUE1
    assert obj.load_foo() == EnumFromScopedModule.VALUE1
    obj.save_foo(EnumFromScopedModule.VALUE2)
    assert obj.foo == EnumFromScopedModule.VALUE2


def test_undefinedAndNull():
    calc = Calculator()
    assert calc.max_value is None
    calc.max_value = None


def test_arrays():
    sum_ = Sum()
    sum_.parts = [Number(5), Number(10), Multiply(Number(2), Number(3))]

    assert sum_.value == 5 + 10 + (2 * 3)
    assert sum_.parts[0].value == 5
    assert sum_.parts[2].value == 6
    assert sum_.to_string() == "(((0 + 5) + 10) + (2 * 3))"


def test_maps():
    calc2 = Calculator()  # Initializer overload (props is optional)
    calc2.add(10)
    calc2.add(20)
    calc2.mul(2)

    assert len(cast(List, calc2.operations_map.get("add"))) == 2
    assert len(cast(List, calc2.operations_map.get("mul"))) == 1
    got = calc2.operations_map.get("add")
    assert got is not None
    assert got[1].value == 30


def test_exceptions():
    calc3 = Calculator(initial_value=20, maximum_value=30)
    calc3.add(3)

    assert calc3.value == 23

    with pytest.raises(RuntimeError):
        calc3.add(10)

    calc3.max_value = 40
    calc3.add(10)

    assert calc3.value == 33


def test_unionProperties():
    calc3 = Calculator()
    calc3.union_property = Multiply(Number(9), Number(3))

    assert isinstance(calc3.union_property, Multiply)
    assert calc3.read_union_value() == 9 * 3

    calc3.union_property = Power(Number(10), Number(3))

    assert isinstance(calc3.union_property, Power)
    assert calc3.read_union_value() == 10**3


def test_subclassing():
    calc = Calculator()
    calc.curr = AddTen(33)
    calc.neg()

    assert calc.value == -43


def test_testJSObjectLiteralToNative():
    obj = JSObjectLiteralToNative()
    obj2 = obj.return_literal()

    assert obj2.prop_a == "Hello"
    assert obj2.prop_b == 102


def test_testFluentApiWithDerivedClasses():
    # make sure that fluent API can be assigned to objects from derived classes
    obj = DerivedFromAllTypes()
    obj.string_property = "Hello"
    obj.number_property = 12

    assert obj.string_property == "Hello"
    assert obj.number_property == 12


def test_creationOfNativeObjectsFromJavaScriptObjects():
    """
    See that we can create a native object, pass it JS and then unmarshal
    back without type information.
    """
    types = AllTypes()

    js_obj = Number(44)
    types.any_property = js_obj
    unmarshalled_js_obj = types.any_property
    assert unmarshalled_js_obj.__class__ == Number

    native_obj = AddTen(10)
    types.any_property = native_obj

    result1 = types.any_property
    assert result1 is native_obj

    native_obj2 = MulTen(20)
    types.any_property = native_obj2
    unmarshalled_native_obj = types.any_property
    assert unmarshalled_native_obj.__class__ == MulTen


def test_asyncOverrides_callAsyncMethod():
    obj = AsyncVirtualMethods()
    assert obj.call_me() == 128
    assert obj.override_me(44) == 528


def test_asyncOverrides_overrideAsyncMethod():
    obj = OverrideAsyncMethods()
    assert obj.call_me() == 4452


def test_asyncOverrides_overrideAsyncMethodByParentClass():
    obj = OverrideAsyncMethodsByBaseClass()
    assert obj.call_me() == 4452


def test_asyncOverrides_overrideCallsSuper():
    obj = OverrideCallsSuper()
    assert obj.override_me(12) == 1441
    assert obj.call_me() == 1209


def test_asyncOverrides_twoOverrides():
    obj = TwoOverrides()
    assert obj.call_me() == 684


def test_asyncOverrides_overrideThrows():
    class ThrowingAsyncVirtualMethods(AsyncVirtualMethods):
        def override_me(self, mult):
            raise RuntimeError("Thrown by native code")

    obj = ThrowingAsyncVirtualMethods()

    with pytest.raises(RuntimeError, match="Thrown by native code"):
        obj.call_me()


def test_syncOverrides():
    obj = SyncOverrides()
    assert obj.caller_is_method() == 10 * 5

    # affect the result
    obj.multiplier = 5
    assert obj.caller_is_method() == 10 * 5 * 5

    # verify callbacks are invoked from a property
    assert obj.caller_is_property == 10 * 5 * 5

    # and from an async method
    obj.multiplier = 3
    assert obj.caller_is_async() == 10 * 5 * 3


def test_propertyOverrides_get_set():
    so = SyncOverrides()
    assert so.retrieve_value_of_the_property() == "I am an override!"
    so.modify_value_of_the_property("New Value")
    assert so.another_the_property == "New Value"


def test_propertyOverrides_get_calls_super():
    class SuperSyncVirtualMethods(SyncVirtualMethods):
        @property
        def the_property(self):
            return f"super:{super().the_property}"

        @the_property.setter
        def the_property(self, value):
            super().the_property = value

    so = SuperSyncVirtualMethods()

    assert so.retrieve_value_of_the_property() == "super:initial value"
    assert so.the_property == "super:initial value"


def test_propertyOverrides_set_calls_super():
    class SuperSyncVirtualMethods(SyncVirtualMethods):
        @property
        def the_property(self):
            return super().the_property

        @the_property.setter
        def the_property(self, value):
            #
            # This is the way this was originally coded:
            #   super().the_property = f"{value}:by override"
            # but this causes a problem because of:
            #   https://bugs.python.org/issue14965
            # so now we have this more convoluted form.
            super(self.__class__, self.__class__).the_property.__set__(  # type: ignore
                self, f"{value}:by override"
            )

    so = SuperSyncVirtualMethods()
    so.modify_value_of_the_property("New Value")

    assert so.the_property == "New Value:by override"


def test_propertyOverrides_get_throws():
    class ThrowingSyncVirtualMethods(SyncVirtualMethods):
        @property
        def the_property(self):
            raise RuntimeError("Oh no, this is bad")

        @the_property.setter
        def the_property(self, value):
            super().the_property = value

    so = ThrowingSyncVirtualMethods()

    with pytest.raises(RuntimeError, match="Oh no, this is bad"):
        so.retrieve_value_of_the_property()


def test_propertyOverrides_set_throws():
    class ThrowingSyncVirtualMethods(SyncVirtualMethods):
        @property
        def the_property(self):
            return super().the_property

        @the_property.setter
        def the_property(self, value):
            raise RuntimeError("Exception from overloaded setter")

    so = ThrowingSyncVirtualMethods()

    with pytest.raises(RuntimeError, match="Exception from overloaded setter"):
        so.modify_value_of_the_property("Hii")


def test_propertyOverrides_interfaces():
    @jsii.implements(IInterfaceWithProperties)
    class TInterfaceWithProperties:
        x = None

        @property
        def read_only_string(self):
            return "READ_ONLY_STRING"

        @property
        def read_write_string(self):
            return f"{self.x}?"

        @read_write_string.setter
        def read_write_string(self, value):
            self.x = f"{value}!"

    obj = TInterfaceWithProperties()
    interact = UsesInterfaceWithProperties(obj)

    assert interact.just_read() == "READ_ONLY_STRING"
    assert interact.write_and_read("Hello") == "Hello!?"


def test_interfaceBuilder():
    @jsii.implements(IInterfaceWithProperties)
    class TInterfaceWithProperties:
        x = "READ_WRITE"

        @property
        def read_only_string(self):
            return "READ_ONLY"

        @property
        def read_write_string(self):
            return self.x

        @read_write_string.setter
        def read_write_string(self, value):
            self.x = value

    obj = TInterfaceWithProperties()
    interact = UsesInterfaceWithProperties(obj)
    assert interact.just_read() == "READ_ONLY"
    assert interact.write_and_read("Hello") == "Hello"


def test_syncOverrides_callsSuper():
    obj = SyncOverrides()
    assert obj.caller_is_property == 10 * 5
    obj.return_super = True
    assert obj.caller_is_property == 10 * 2


def test_fail_syncOverrides_callsDoubleAsync_method():
    obj = SyncOverrides()
    obj.call_async = True

    # TODO: Error Handling
    with pytest.raises(RuntimeError):
        obj.caller_is_method()


def test_fail_syncOverrides_callsDoubleAsync_propertyGetter():
    obj = SyncOverrides()
    obj.call_async = True

    # TODO: Error Handling
    with pytest.raises(RuntimeError):
        obj.caller_is_property


@pytest.mark.skip
def test_fail_syncOverrides_callsDoubleAsync_propertySetter():
    obj = SyncOverrides()
    obj.call_async = True

    # TODO: Error Handling
    with pytest.raises(RuntimeError):
        obj.caller_is_property = 12


def test_testInterfaces() -> None:
    friendly: IFriendly
    friendlier: IFriendlier
    random_number_generator: IRandomNumberGenerator
    friendly_random_generator: IFriendlyRandomGenerator

    add = Add(Number(10), Number(20))
    friendly = add
    assert friendly.hello() == "Hello, I am a binary operation. What's your name?"

    multiply = Multiply(Number(10), Number(30))
    friendly = multiply
    friendlier = multiply
    random_number_generator = multiply
    assert friendly.hello() == "Hello, I am a binary operation. What's your name?"
    assert friendlier.goodbye() == "Goodbye from Multiply!"
    assert random_number_generator.next() == 89

    friendly_random_generator = DoubleTrouble()
    assert friendly_random_generator.hello() == "world"
    assert friendly_random_generator.next() == 12

    poly = Polymorphism()
    assert (
        poly.say_hello(friendly)
        == "oh, Hello, I am a binary operation. What's your name?"
    )
    assert poly.say_hello(friendly_random_generator) == "oh, world"
    assert (
        poly.say_hello(SubclassNativeFriendlyRandom())
        == "oh, SubclassNativeFriendlyRandom"
    )
    assert poly.say_hello(PureNativeFriendlyRandom()) == "oh, I am a native!"


def test_testNativeObjectsWithInterfaces():
    # create a pure and native object, not part of the jsii hierarchy, only implements
    # a jsii interface
    pure_native = PureNativeFriendlyRandom()
    subclassed_native = SubclassNativeFriendlyRandom()
    generator_bound_to_p_subclassed_object = NumberGenerator(subclassed_native)
    generator_bound_to_pure_native = NumberGenerator(pure_native)

    assert generator_bound_to_p_subclassed_object.generator is subclassed_native
    generator_bound_to_p_subclassed_object.is_same_generator(subclassed_native)
    assert generator_bound_to_p_subclassed_object.next_times100() == 10000

    # When we invoke nextTimes100 again, it will use the objref and call into the same
    # object.
    assert generator_bound_to_p_subclassed_object.next_times100() == 20000

    assert generator_bound_to_pure_native.generator is pure_native
    generator_bound_to_pure_native.is_same_generator(pure_native)
    assert generator_bound_to_pure_native.next_times100() == 100_000
    assert generator_bound_to_pure_native.next_times100() == 200_000


def test_testLiteralInterface():
    obj = JSObjectLiteralForInterface()
    friendly = obj.give_me_friendly()
    gen = obj.give_me_friendly_generator()

    assert friendly.hello() == "I am literally friendly!"
    assert gen.hello() == "giveMeFriendlyGenerator"
    assert gen.next() == 42


def test_testInterfaceParameter():
    obj = JSObjectLiteralForInterface()
    friendly = obj.give_me_friendly()
    greeting_augmenter = GreetingAugmenter()

    assert friendly.hello() == "I am literally friendly!"
    assert (
        greeting_augmenter.better_greeting(friendly)
        == "I am literally friendly! Let me buy you a drink!"
    )


def test_statics():
    assert Statics.static_method("Yoyo") == "hello ,Yoyo!"
    assert Statics.instance.value == "default"

    new_statics = Statics("new value")
    Statics.instance = new_statics

    assert Statics.instance is new_statics
    assert Statics.instance.value == "new value"

    assert Statics.non_const_static == 100


def test_consts():
    obj = Statics.CONST_OBJ

    assert Statics.FOO == "hello"
    assert obj.hello() == "world"
    assert Statics.BAR == 1234
    assert Statics.ZOO_BAR.get("hello") == "world"


def test_reservedKeywordsAreSlugifiedInMethodNames():
    obj = PythonReservedWords()
    obj.import_()
    obj.return_()


def test_nodeStandardLibrary():
    obj = NodeStandardLibrary()

    assert obj.fs_read_file() == "Hello, resource!"
    assert obj.fs_read_file_sync() == "Hello, resource! SYNC!"
    assert len(obj.os_platform) > 0
    assert (
        obj.crypto_sha256()
        == "6a2da20943931e9834fc12cfe5bb47bbd9ae43489a30726962b576f4e3993e50"
    )


def test_returnAbstract():
    obj = AbstractClassReturner()
    obj2 = obj.give_me_abstract()

    assert obj2.abstract_method("John") == "Hello, John!!"
    assert obj2.prop_from_interface == "propFromInterfaceValue"
    assert obj2.non_abstract_method() == 42

    iface = obj.give_me_interface()
    assert iface.prop_from_interface == "propFromInterfaceValue"

    assert (
        obj.return_abstract_from_property.abstract_property == "hello-abstract-property"
    )


def test_doNotOverridePrivates_method():
    class TDoNotOverridePrivates(DoNotOverridePrivates):
        def private_method(self):
            return "privateMethod-Override"

    obj = TDoNotOverridePrivates()

    assert obj.private_method_value() == "privateMethod"


def test_doNotOverridePrivates_property_by_name():
    class TDoNotOverridePrivates(DoNotOverridePrivates):
        def private_property(self):
            return "privateProperty-Override"

    obj = TDoNotOverridePrivates()

    assert obj.private_property_value() == "privateProperty"


def test_doNotOverridePrivates_property_getter():
    class TDoNotOverridePrivates(DoNotOverridePrivates):
        @property
        def private_property(self):
            return "privateProperty-Override"

        @private_property.setter
        def private_property(self):
            raise RuntimeError("Boom")

    obj = TDoNotOverridePrivates()

    assert obj.private_property_value() == "privateProperty"

    # verify the setter override is not invoked.
    obj.change_private_property_value("MyNewValue")
    assert obj.private_property_value() == "MyNewValue"


def test_classWithPrivateConstructorAndAutomaticProperties():
    obj = ClassWithPrivateConstructorAndAutomaticProperties.create("Hello", "Bye")
    assert obj.read_write_string == "Bye"
    assert obj.read_only_string == "Hello"


def test_nullShouldBeTreatedAsUndefined():
    obj = NullShouldBeTreatedAsUndefined("hello", None)
    obj.give_me_undefined(None)
    obj.give_me_undefined_inside_an_object(
        this_should_be_undefined=None,
        array_with_three_elements_and_undefined_as_second_argument=[
            "hello",
            None,
            "boom",
        ],
    )
    obj.change_me_to_undefined = None
    obj.verify_property_is_undefined()


def test_testJsiiAgent():
    assert JsiiAgent.value == f"Python/{platform.python_version()}"


def test_receiveInstanceOfPrivateClass():
    assert ReturnsPrivateImplementationOfInterface().private_implementation.success


def test_eraseUnsetDataValues():
    opts = EraseUndefinedHashValuesOptions(option1="option1")
    assert EraseUndefinedHashValues.does_key_exist(opts, "option1")
    assert not EraseUndefinedHashValues.does_key_exist(opts, "option2")


def test_objectIdDoesNotGetReallocatedWhenTheConstructorPassesThisOut():
    class PartiallyInitializedThisConsumerImpl(PartiallyInitializedThisConsumer):
        def consume_partially_initialized_this(self, obj, dt, en):
            assert obj is not None
            assert isinstance(dt, datetime)
            assert en == AllTypesEnum.THIS_IS_GREAT
            return "OK"

    reflector = PartiallyInitializedThisConsumerImpl()
    obj = ConstructorPassesThisOut(reflector)
    assert obj is not None


def test_variadicMethodCanBeInvoked():
    variadic = VariadicMethod(1)
    assert variadic.as_array(3, 4, 5, 6) == [1, 3, 4, 5, 6]


def test_callbacksCorrectlyDeserializeArguments():
    class DataRendererSubclass(DataRenderer):
        def render_map(self, map):
            return super().render_map(map)

    renderer = DataRendererSubclass()
    assert (
        renderer.render(anumber=42, astring="bazinga!")
        == '{\n  "anumber": 42,\n  "astring": "bazinga!"\n}'
    )


def test_passNestedStruct():
    output = StructPassing.round_trip(
        123,
        required="hello",
        second_level=SecondLevelStruct(deeper_required_prop="exists"),
    )

    assert output.required == "hello"
    assert output.optional is None
    assert cast(SecondLevelStruct, output.second_level).deeper_required_prop == "exists"

    # Test stringification
    # Dicts are ordered in Python 3.7+, so this is fine: https://mail.python.org/pipermail/python-dev/2017-December/151283.html
    assert (
        str(output)
        == "TopLevelStruct(required='hello', second_level=SecondLevelStruct(deeper_required_prop='exists'))"
    )


def test_passNestedScalar():
    output = StructPassing.round_trip(123, required="hello", second_level=5)

    assert output.required == "hello"
    assert output.optional is None
    assert output.second_level == 5


def test_passStructsInVariadic():
    output = StructPassing.how_many_var_args_did_i_pass(
        123,
        TopLevelStruct(required="hello", second_level=1),
        TopLevelStruct(
            required="bye", second_level=SecondLevelStruct(deeper_required_prop="ciao")
        ),
    )
    assert output == 2


def test_structEquality():
    a = TopLevelStruct(
        required="bye", second_level=SecondLevelStruct(deeper_required_prop="ciao")
    )
    b = (TopLevelStruct(required="hello", second_level=1),)
    c = (TopLevelStruct(required="hello", second_level=1),)
    d = SecondLevelStruct(deeper_required_prop="exists")

    assert a != b
    assert b == c
    assert a != 5
    assert a != d


def test_correctly_handling_struct_unions():
    a0 = StructA(required_string="Present!", optional_string="Bazinga!")
    a1 = StructA(required_string="Present!", optional_number=1337)
    b0 = StructB(required_string="Present!", optional_boolean=True)
    b1 = StructB(required_string="Present!", optional_struct_a=a1)

    assert StructUnionConsumer.is_struct_a(a0)
    assert StructUnionConsumer.is_struct_a(a1)
    assert not StructUnionConsumer.is_struct_a(b0)
    assert not StructUnionConsumer.is_struct_a(b1)

    assert not StructUnionConsumer.is_struct_b(a0)
    assert not StructUnionConsumer.is_struct_b(a1)
    assert StructUnionConsumer.is_struct_b(b0)
    assert StructUnionConsumer.is_struct_b(b1)


def test_consumer_calls_method_static_objliteral():
    assert ConsumerCanRingBell.static_implemented_by_object_literal(PythonBellRinger())


def test_consumer_calls_method_static_publicclass():
    assert ConsumerCanRingBell.static_implemented_by_public_class(PythonBellRinger())


def test_consumer_calls_method_static_privateclass():
    assert ConsumerCanRingBell.static_implemented_by_private_class(PythonBellRinger())


def test_consumer_calls_method_static_typed_as_class():
    assert ConsumerCanRingBell.static_when_typed_as_class(PythonConcreteBellRinger())


def test_consumer_calls_method_objliteral():
    assert ConsumerCanRingBell().implemented_by_object_literal(PythonBellRinger())


def test_consumer_calls_method_publicclass():
    assert ConsumerCanRingBell().implemented_by_public_class(PythonBellRinger())


def test_consumer_calls_method_privateclass():
    assert ConsumerCanRingBell().implemented_by_private_class(PythonBellRinger())


def test_consumer_calls_method_typed_as_class():
    assert ConsumerCanRingBell().when_typed_as_class(PythonConcreteBellRinger())


def test_can_pass_nested_struct_as_dict():
    # Those shouldn't raise:
    RootStructValidator.validate(string_prop="Pickle Rick!!!")
    RootStructValidator.validate(string_prop="Pickle Rick!!!", nested_struct=None)
    RootStructValidator.validate(
        string_prop="Pickle Rick!!!", nested_struct={"number_prop": 1337}
    )


def test_can_leverage_indirect_interface_polymorphism():
    provider = AnonymousImplementationProvider()
    assert provider.provide_as_class().value == 1337
    assert provider.provide_as_interface().value == 1337
    assert provider.provide_as_interface().verb() == "to implement"


# https://github.com/aws/jsii/issues/976
def test_return_subclass_that_implements_interface_976():
    obj = SomeTypeJsii976.return_return()
    assert obj.foo == 333


def test_return_subclass_that_implements_interface_976_raises_attributeerror_when_using_non_existent_method():
    obj = SomeTypeJsii976.return_return()
    try:
        print(obj.not_a_real_method_I_swear)  # type:ignore
        failed = False
    except AttributeError as err:
        failed = True
        assert (
            err.args[0]
            == "'<class 'jsii_calc.BaseJsii976'>+<class 'jsii_calc._IReturnJsii976Proxy'>' object has no attribute 'not_a_real_method_I_swear'"
        )
    assert failed


def test_return_anonymous_implementation_of_interface():
    assert SomeTypeJsii976.return_anonymous() is not None


def test_structs_can_be_downcasted_to_parent_type():
    assert Demonstrate982.take_this() is not None
    assert Demonstrate982.take_this_too() is not None


@jsii.implements(IBellRinger)
class PythonBellRinger:
    def your_turn(self, bell):
        bell.ring()


@jsii.implements(IConcreteBellRinger)
class PythonConcreteBellRinger:
    def your_turn(self, bell):
        bell.ring()


def test_null_is_a_valid_optional_list():
    assert DisappointingCollectionSource.MAYBE_LIST is None


def test_null_is_a_valid_optional_map():
    assert DisappointingCollectionSource.MAYBE_MAP is None


def test_can_use_interface_setters():
    obj = ObjectWithPropertyProvider.provide()
    obj.property = "New Value"
    assert obj.was_set()


def test_structs_are_undecorated_on_the_way_to_kernel():
    json = JsonFormatter.stringify(
        StructB(required_string="Bazinga!", optional_boolean=False)
    )
    assert loads(cast(str, json)) == {
        "requiredString": "Bazinga!",
        "optionalBoolean": False,
    }


def test_can_obtain_reference_with_overloaded_setter():
    assert ConfusingToJackson.make_instance() is not None


def test_can_obtain_struct_reference_with_overloaded_setter():
    assert ConfusingToJackson.make_struct_instance() is not None


def test_pure_interfaces_can_be_used_transparently():
    expected = StructB(required_string="It's Britney b**ch!")

    @jsii.implements(IStructReturningDelegate)
    class StructReturningDelegate:
        def return_struct(self):
            return expected

    delegate = StructReturningDelegate()
    consumer = ConsumePureInterface(delegate)
    assert consumer.work_it_baby() == expected


def test_pure_interfaces_can_be_used_transparently_when_transitively_implementing():
    expected = StructB(required_string="It's Britney b**ch!")

    @jsii.implements(IStructReturningDelegate)
    class ImplementsStructReturningDelegate:
        def return_struct(self):
            return expected

    class IndirectlyImplementsStructReturningDelegate(
        ImplementsStructReturningDelegate
    ):
        ...

    delegate = IndirectlyImplementsStructReturningDelegate()
    consumer = ConsumePureInterface(delegate)
    assert consumer.work_it_baby() == expected


def test_pure_interfaces_can_be_used_transparently_when_added_to_jsii_type():
    expected = StructB(required_string="It's Britney b**ch!")

    @jsii.implements(IStructReturningDelegate)
    class ImplementsAdditionalInterface(AllTypes):
        def return_struct(self):
            return expected

    delegate = ImplementsAdditionalInterface()
    consumer = ConsumePureInterface(delegate)
    assert consumer.work_it_baby() == expected


def test_lifted_kwarg_with_same_name_as_positional_arg():
    bell = Bell()
    amb = AmbiguousParameters(bell, scope="Driiiing!")

    assert amb.scope == bell
    assert amb.props == StructParameterType(scope="Driiiing!")


def test_abstract_members_are_correctly_handled():
    class AbstractSuiteImpl(AbstractSuite):
        @property
        def _property(self):
            return self.property

        @_property.setter
        def _property(self, value):
            self.property = "String<%s>" % value

        def _some_method(self, str):
            return "Wrapped<%s>" % str

    abstract_suite = AbstractSuiteImpl()
    assert "Wrapped<String<Oomf!>>" == abstract_suite.work_it_all("Oomf!")


def test_collection_of_interfaces_list_of_structs():
    for elt in InterfaceCollections.list_of_structs():
        assert getattr(elt, "required_string") is not None


def test_collection_of_interfaces_list_of_interfaces():
    for elt in InterfaceCollections.list_of_interfaces():
        assert getattr(elt, "ring") is not None


def test_collection_of_interfaces_map_of_structs():
    for elt in InterfaceCollections.map_of_structs().values():
        assert getattr(elt, "required_string") is not None


def test_collection_of_interfaces_map_of_interfaces():
    for elt in InterfaceCollections.map_of_interfaces().values():
        assert getattr(elt, "ring") is not None


def test_dependency_submodule_types_are_usable():
    subject = UpcasingReflectable({"foo": "bar"})

    assert UpcasingReflectable.REFLECTOR.as_map(subject) == {"FOO": "bar"}


def test_load_submodules():
    from jsii_calc.submodule import nested_submodule
    import jsii_calc.submodule


def test_parameter_named_self_ClassWithSelf():
    subject = ClassWithSelf("Howdy!")
    assert subject.self == "Howdy!"
    assert subject.method(1337) == "1337"


def test_parameter_named_self_ClassWithSelfKwarg():
    subject = ClassWithSelfKwarg(self="Howdy!")
    assert subject.props.self == "Howdy!"


def test_isomorphism_within_constructor():
    class Subject(Isomorphism):
        def __init__(self):
            super().__init__()
            assert self == self.myself()

    Subject()


def test_kwargs_from_superinterface_are_working():
    assert Kwargs.method(extra="ordinary", prop=SomeEnum.SOME)


def test_iso8601_does_not_deserialize_to_date():
    @jsii.implements(IWallClock)
    class WallClock:
        def __init__(self, now: str):
            self.now = now

        def iso8601_now(self) -> str:
            return self.now

    class MildEntropy(Entropy):
        def repeat(self, word: str) -> str:
            return word

    now = datetime.utcnow().isoformat() + "Z"
    wall_clock = WallClock(now)
    entropy = MildEntropy(wall_clock)

    assert now == entropy.increase()


def test_class_can_extend_and_implement_from_jsii():
    """
    This test is identical to test_iso8601_does_not_deserialize_to_date, except
    the WallCloc class extends ClassWithSelf (a well-known jsii type), to
    demonstrate it is possible to both extend a jsii type, and implement a
    supplemental interface at the same time.

    See also https://github.com/aws/jsii/issues/2963
    """

    @jsii.implements(IWallClock)
    class WallClock(ClassWithSelf):
        def __init__(self, now: str):
            super().__init__(now)
            self.now = now

        def iso8601_now(self) -> str:
            return self.now

    class MildEntropy(Entropy):
        def repeat(self, word: str) -> str:
            return word

    now = datetime.utcnow().isoformat() + "Z"
    wall_clock = WallClock(now)
    entropy = MildEntropy(wall_clock)

    assert now == entropy.increase()


def test_class_can_be_used_when_not_expressedly_loaded():
    """
    This test verifies that it is possible to receive instances of classes that
    belong to submodules that have not been explicitly imported. This implies
    the types are registered with the runtime even though the submodule has
    never been mentioned in userland.
    """

    class Subject(Cdk16625):
        def _unwrap(self, rng: IRandomNumberGenerator):
            return rng.next()

    # This should NOT throw
    Subject().test()


def test_stripped_deprecated_member_can_be_received():
    assert InterfaceFactory.create() is not None


def test_exception_message():
    with pytest.raises(RuntimeError, match="Cannot find asset"):
        AcceptsPath(source_path="A Bad Path")


def test_void_returning_async():
    """Verifies it's okay to return a Promise<void>."""

    assert PromiseNothing().instance_promise_it() is None
    ## TODO: This is currently broken as code-gen is incorrect for static async.
    # assert PromiseNothing.promise_it() is None



from typing import Any, Callable, Optional, Type, Union


class _ClassProperty:
    def __init__(
        self,
        fget: classmethod,
        fset: Optional[classmethod] = None,
    ):
        self.fget = fget
        self.fset = fset

    def __get__(self, obj: Any, klass: Optional[Type] = None) -> Any:
        if klass is None:
            klass = type(obj)
        return self.fget.__get__(obj, klass)()

    def __set__(self, obj: Any, value: Any) -> None:
        if self.fset is None:
            raise AttributeError("Can't set class property (no setter)")
        klass = type(obj)
        return self.fset.__get__(obj, klass)(value)

    def setter(
        self, fset: Union[Callable[[Any, Any], None], classmethod]
    ) -> "_ClassProperty":
        """
        Defines the setter for a class property
        """
        if not isinstance(fset, classmethod):
            fset = classmethod(fset)
        self.fset = fset
        return self


def classproperty(fget: Union[Callable[[Any], Any], classmethod]) -> _ClassProperty:
    """
    Declares a new class property with the decorated getter.
    """
    if not isinstance(fget, classmethod):
        fget = classmethod(fget)
    return _ClassProperty(fget)


class _ClassPropertyMeta(type):
    def __setattr__(self, key: str, value: Any) -> None:
        obj = getattr(self, key, None)
        if isinstance(obj, _ClassProperty):
            return obj.__set__(self, value)

        return super().__setattr__(key, value)



from typing import Any, Dict, Generic, List, Optional, Mapping, TypeVar, Union
from typing_extensions import Protocol

import attr


@attr.s(auto_attribs=True, frozen=True, slots=True)
class ObjRef:
    ref: str
    interfaces: Optional[List[str]] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class EnumRef:
    ref: ObjRef
    member: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Override:
    method: Optional[str] = None
    property: Optional[str] = None
    cookie: Optional[str] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class LoadRequest:
    name: str
    version: str
    tarball: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class LoadResponse:
    assembly: str
    types: int


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GetScriptCommandRequest:
    assembly: str
    script: str
    args: List[Any] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GetScriptCommandResponse:
    command: str
    args: List[str] = attr.Factory(list)
    env: Dict[str, str] = attr.Factory(dict)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class InvokeScriptRequest:
    assembly: str
    script: str
    args: List[Any] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class InvokeScriptResponse:
    status: int
    stdout: str
    stderr: str
    signal: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CreateRequest:
    fqn: str
    args: List[Any] = attr.Factory(list)
    overrides: List[Override] = attr.Factory(list)
    interfaces: Optional[List[str]] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CreateResponse(ObjRef):
    ...


@attr.s(auto_attribs=True, frozen=True, slots=True)
class DeleteRequest:
    objref: ObjRef


@attr.s(auto_attribs=True, frozen=True, slots=True)
class DeleteResponse:
    ...


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GetRequest:
    objref: ObjRef
    property: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StaticGetRequest:
    fqn: str
    property: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class GetResponse:
    value: Any = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StaticSetRequest:
    fqn: str
    property: str
    value: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class SetRequest:
    objref: ObjRef
    property: str
    value: Any


@attr.s(auto_attribs=True, frozen=True, slots=True)
class SetResponse:
    ...


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StaticInvokeRequest:
    fqn: str
    method: str
    args: Optional[List[Any]] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class InvokeRequest:
    objref: ObjRef
    method: str
    args: Optional[List[Any]] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class InvokeResponse:
    result: Any = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class BeginRequest:
    objref: ObjRef
    method: str
    args: Optional[List[Any]] = attr.Factory(list)


@attr.s(auto_attribs=True, frozen=True, slots=True)
class BeginResponse:
    promiseid: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class EndRequest:
    promiseid: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class EndResponse:
    result: Optional[Any] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class Callback:
    cbid: str
    cookie: str
    invoke: Optional[InvokeRequest] = None
    get: Optional[GetRequest] = None
    set: Optional[SetRequest] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CallbacksRequest:
    ...


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CallbacksResponse:
    callbacks: List[Callback]


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CompleteRequest:
    cbid: str
    err: Optional[str] = None
    result: Optional[Any] = None


@attr.s(auto_attribs=True, frozen=True, slots=True)
class CompleteResponse:
    cbid: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class NamingRequest:
    assembly: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class NamingResponse:
    naming: Mapping[str, Mapping[str, Optional[Any]]]


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StatsRequest:
    ...


@attr.s(auto_attribs=True, frozen=True, slots=True)
class StatsResponse:
    objectCount: int


KernelRequest = Union[
    LoadRequest,
    CreateRequest,
    DeleteRequest,
    GetRequest,
    SetRequest,
    StaticGetRequest,
    InvokeRequest,
    InvokeScriptRequest,
    StaticInvokeRequest,
    StatsRequest,
]

KernelResponse = Union[
    BeginResponse,
    LoadResponse,
    CreateResponse,
    DeleteResponse,
    GetResponse,
    InvokeResponse,
    InvokeScriptResponse,
    SetResponse,
    StatsResponse,
    Callback,
]



EMBEDDED_FILES = {
    "bin__jsii-runtime.js": "bin/jsii-runtime.js",
    "bin__jsii-runtime.js.map": "bin/jsii-runtime.js.map",
    "lib__program.js": "lib/program.js",
    "lib__program.js.map": "lib/program.js.map",
}

ENTRYPOINT = "bin__jsii-runtime.js"

__all__ = ["EMBEDDED_FILES", "ENTRYPOINT"]



import abc
import os
import sys
import subprocess

import attr

from typing import (
    Any,
    Callable,
    cast,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    TypeVar,
)

from . import _reference_map
from ._compat import importlib_resources
from ._kernel import Kernel
from .python import _ClassPropertyMeta


# Yea, a global here is kind of gross, however, there's not really a better way of
# handling this. Fundamentally this is a global value, since we can only reasonably
# have a single kernel active at any one time in a real program.
kernel = Kernel()


@attr.s(auto_attribs=True, frozen=True, slots=True)
class JSIIAssembly:
    name: str
    version: str
    module: str
    filename: str

    @classmethod
    def load(cls, *args, _kernel=kernel, **kwargs) -> "JSIIAssembly":
        # Our object here really just acts as a record for our JSIIAssembly, it doesn't
        # offer any functionality itself, besides this class method that will trigger
        # the loading of the given assembly in the JSII Kernel.
        assembly = cls(*args, **kwargs)

        # Actually load the assembly into the kernel, we're using the
        # importlib.resources API here instead of manually constructing the path, in
        # the hopes that this will make JSII modules able to be used with zipimport
        # instead of only on the FS.
        with importlib_resources.as_file(
            importlib_resources.files(f"{assembly.module}._jsii").joinpath(
                assembly.filename
            )
        ) as assembly_path:
            _kernel.load(assembly.name, assembly.version, os.fspath(assembly_path))

        # Give our record of the assembly back to the caller.
        return assembly

    @classmethod
    def invokeBinScript(
        cls,
        pkgname: str,
        script: str,
        args: Optional[Sequence[str]] = None,
        _kernel=kernel,
    ) -> int:
        if args is None:
            args = []

        response = _kernel.getBinScriptCommand(pkgname, script, args)

        result = subprocess.run(
            " ".join([response.command, *response.args]),
            encoding="utf-8",
            shell=True,
            env=response.env,
        )
        return result.returncode


M = TypeVar("M")


class JSIIMeta(_ClassPropertyMeta, type):
    def __new__(
        cls: Type["JSIIMeta"],
        name: str,
        bases: tuple,
        attrs: dict,
        *,
        jsii_type: Optional[str] = None,
    ) -> "JSIIMeta":
        # We want to ensure that subclasses of a JSII class do not require setting the
        # jsii_type keyword argument. They should be able to subclass it as normal.
        # Since their parent class will have the __jsii_type__ variable defined, they
        # will as well anyways.
        if jsii_type is not None:
            attrs["__jsii_type__"] = jsii_type
        # The declared type should NOT be inherited by subclasses. This way we can identify whether
        # an MRO entry corresponds to a possible overrides contributor or not.
        attrs["__jsii_declared_type__"] = jsii_type

        obj = super().__new__(cls, name, bases, attrs)

        # Now that we've created the class, we'll need to register it with our reference
        # mapper. We only do this for types that are actually jsii types, and not any
        # subclasses of them.
        if jsii_type is not None:
            _reference_map.register_type(obj)

        return cast("JSIIMeta", obj)

    def __call__(cls: Type[M], *args: Any, **kwargs) -> M:
        # There is no way to constrain the metaclass of a `Type[M]` hint today, so we have to
        # perform a `cast` trick here in order for MyPy to accept this code as valid... The implicit
        # arguments to `super()` otherwise are `super(__class__, cls)`, which results in an error.
        inst = super(JSIIMeta, cast(JSIIMeta, cls)).__call__(*args, **kwargs)

        # Register this instance with our reference map.
        _reference_map.register_reference(inst)

        return inst


class JSIIAbstractClass(abc.ABCMeta, JSIIMeta):
    pass


F = TypeVar("F", bound=Callable[..., Any])
T = TypeVar("T", bound=Type[Any])


def enum(*, jsii_type: str) -> Callable[[T], T]:
    def deco(cls):
        cls.__jsii_type__ = jsii_type
        _reference_map.register_enum(cls)
        return cls

    return deco


def data_type(
    *,
    jsii_type: str,
    jsii_struct_bases: List[Type[Any]],
    name_mapping: Mapping[str, str],
) -> Callable[[T], T]:
    def deco(cls):
        cls.__jsii_type__ = jsii_type
        cls.__jsii_struct_bases__ = jsii_struct_bases
        cls.__jsii_name_mapping__ = name_mapping
        _reference_map.register_data_type(cls)
        return cls

    return deco


def member(*, jsii_name: str) -> Callable[[F], F]:
    def deco(fn):
        fn.__jsii_name__ = jsii_name
        return fn

    return deco


def implements(*interfaces: Type[Any]) -> Callable[[T], T]:
    def deco(cls):
        cls.__jsii_type__ = getattr(cls, "__jsii_type__", None)
        cls.__jsii_ifaces__ = getattr(cls, "__jsii_ifaces__", []) + list(interfaces)
        return cls

    return deco


def interface(*, jsii_type: str) -> Callable[[T], T]:
    def deco(iface):
        iface.__jsii_type__ = jsii_type
        _reference_map.register_interface(iface)
        return iface

    return deco


def proxy_for(abstract_class: Type[Any]) -> Type[Any]:
    if not hasattr(abstract_class, "__jsii_proxy_class__"):
        raise TypeError(f"{abstract_class} is not a JSII Abstract class.")

    return cast(Any, abstract_class).__jsii_proxy_class__()


def python_jsii_mapping(cls: Type[Any]) -> Optional[Mapping[str, str]]:
    return getattr(cls, "__jsii_name_mapping__", None)



import sys
from typing import Union

from .__meta__ import __version__, __jsii_runtime_version__
from ._runtime import (
    JSIIAssembly,
    JSIIMeta,
    JSIIAbstractClass,
    enum,
    data_type,
    implements,
    interface,
    member,
    kernel,
    proxy_for,
)
from . import python


# JS doesn't have distinct float or integer types, but we do. So we'll define our own
# type alias that will allow either.
Number = Union[int, float]


# Alias our Kernel methods here, so that jsii.<method> works. This will hide the fact
# that there is a kernel at all from our callers.
load = kernel.load
create = kernel.create
delete = kernel.delete
get = kernel.get
set = kernel.set
sget = kernel.sget
sset = kernel.sset
invoke = kernel.invoke
ainvoke = kernel.ainvoke
sinvoke = kernel.sinvoke
stats = kernel.stats


if sys.version_info < (3, 8):
    from platform import python_version
    import warnings

    warnings.warn(
        f"WARNING: You are using python release {python_version()}, which has reached end-of-life! Support for EOL Python releases may be dropped in the future. Please consider upgrading to Python >= 3.8 as soon as possible.",
    )


__all__ = [
    "__version__",
    "__jsii_runtime_version__",
    "JSIIAssembly",
    "JSIIMeta",
    "JSIIAbstractClass",
    "Number",
    "enum",
    "data_type",
    "implements",
    "interface",
    "member",
    "kernel",
    "proxy_for",
    "load",
    "create",
    "delete",
    "get",
    "set",
    "sget",
    "sset",
    "invoke",
    "ainvoke",
    "sinvoke",
    "stats",
    "python",
]



from os import environ
import platform
import subprocess
from datetime import datetime
import pytest


@pytest.fixture()
def silence_node_deprecation_warnings():
    """Ensures @jsii/check-node does not emit any warning that would cause the
    test output assertions to be invalidated.
    """

    variables = [
        "JSII_SILENCE_WARNING_KNOWN_BROKEN_NODE_VERSION",
        "JSII_SILENCE_WARNING_UNTESTED_NODE_VERSION",
        "JSII_SILENCE_WARNING_DEPRECATED_NODE_VERSION",
        "JSII_SILENCE_WARNING_END_OF_LIFE_NODE_VERSION",
    ]

    store = {var: environ.get(var, "") for var in variables}

    for var in variables:
        environ[var] = "1"

    nodeEolAcknowledgement = "Node14 is now end of life (EOL) as of April 30, 2023. Support of EOL runtimes are only guaranteed for 30 days after EOL. By silencing this warning you acknowledge that you are using an EOL version of Node and will upgrade to a supported version as soon as possible."
    environ["JSII_SILENCE_WARNING_END_OF_LIFE_NODE_VERSION"] = nodeEolAcknowledgement

    # Execute the test
    yield

    for var in variables:
        environ[var] = store[var]


class TestInvokeBinScript:
    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="jsii-pacmak does not generate windows scripts",
    )
    def test_invoke_script(self, silence_node_deprecation_warnings) -> None:
        script_path = f".env/bin/calc"
        result = subprocess.run([script_path], capture_output=True)

        assert result.returncode == 0
        assert result.stdout == b"Hello World!\n"

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="jsii-pacmak does not generate windows scripts",
    )
    def test_invoke_script_with_args(self, silence_node_deprecation_warnings) -> None:
        script_path = f".env/bin/calc"
        result = subprocess.run([script_path, "arg1", "arg2"], capture_output=True)

        assert result.returncode == 0
        assert result.stdout == b"Hello World!\n  arguments: arg1, arg2\n"

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="jsii-pacmak does not generate windows scripts",
    )
    def test_invoke_script_with_failure(
        self, silence_node_deprecation_warnings
    ) -> None:
        script_path = f".env/bin/calc"
        result = subprocess.run([script_path, "arg1", "fail"], capture_output=True)

        assert result.returncode == 3
        assert result.stdout == b"Hello World!\n  arguments: arg1, fail\n"
        assert result.stderr.startswith(b"error message to stderr\n")

    @pytest.mark.skipif(
        platform.system() == "Windows",
        reason="jsii-pacmak does not generate windows scripts",
    )
    def test_invoke_script_with_line_flush(
        self, silence_node_deprecation_warnings
    ) -> None:
        """Make sure lines are flushed immediately as they are generated, rather than
        buffered to the end
        """
        script_path = f".env/bin/calc"
        proc = subprocess.Popen(
            [script_path, "arg1", "delay"], stdout=subprocess.PIPE, bufsize=1
        )

        timed_lines = []
        last_dt = None
        if proc.stdout is not None:
            for line in iter(proc.stdout.readline, b""):
                last_dt = datetime.utcnow() if last_dt is None else last_dt
                current_dt = datetime.utcnow()

                delay_s = round((current_dt - last_dt).total_seconds())
                timed_lines.append([line, delay_s])

                last_dt = current_dt
        exit_code = proc.wait()

        assert exit_code == 0
        assert timed_lines == [
            [b"Hello World!\n", 0],
            [b"  arguments: arg1, delay\n", 0],
            [b"sleeping 1s 1\n", 0],
            [b"sleeping 1s 2\n", 1],
            [b"sleeping 1s 3\n", 1],
            [b"sleeping 1s 4\n", 1],
            [b"sleeping 1s 5\n", 1],
        ]



import functools

from typing import Any, MutableMapping, Type


class Singleton(type):
    _instances: MutableMapping[Type[Any], Any] = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)

        return cls._instances[cls]


def memoized_property(fgetter):
    stored = []

    @functools.wraps(fgetter)
    def wrapped(self):
        nonlocal stored
        if not stored:
            stored.append(fgetter(self))
        return stored[0]

    return property(wrapped)



# This module exists to break an import cycle between jsii.runtime and jsii.kernel
import inspect

from typing import Any, Iterable, List, Mapping, MutableMapping, Type
from ._kernel.types import ObjRef


_types = {}
_data_types: MutableMapping[str, Any] = {}
_enums: MutableMapping[str, Any] = {}
_interfaces: MutableMapping[str, Any] = {}


def register_type(klass: Type):
    _types[klass.__jsii_type__] = klass


def register_data_type(data_type: Any):
    _data_types[data_type.__jsii_type__] = data_type


def register_enum(enum_type: Any):
    _enums[enum_type.__jsii_type__] = enum_type


def register_interface(iface: Any):
    _interfaces[iface.__jsii_type__] = iface


class _FakeReference:
    def __init__(self, ref: str) -> None:
        self.__jsii_ref__ = ref


class _ReferenceMap:
    def __init__(self, types: Mapping[str, Type]) -> None:
        # We are using a real dictionary here instead of a WeakValueDictionary because
        # the nature of the JSII is such that we can never free the memory of JSII
        # objects ever, because we have no idea how many references exist on the *other*
        # side.
        self._refs: MutableMapping[str, Any] = {}
        self._types = types

    def register(self, inst: Any) -> None:
        self._refs[inst.__jsii_ref__.ref] = inst

    def resolve(self, kernel, ref):
        # First we need to check our reference map to see if we have any instance that
        # already matches this reference.
        try:
            # TODO: Handle discovery of possible new interfaces on the ObjRef
            return self._refs[ref.ref]
        except KeyError:
            pass

        # If we got to this point, then we didn't have a referene for this, in that case
        # we want to create a new instance, but we need to create it in such a way that
        # we don't try to recreate the type inside of the JSII interface.
        class_fqn = ref.ref.rsplit("@", 1)[0]
        if class_fqn in _types:
            klass = _types[class_fqn]

            # If this class is an abstract class, then we'll use the generated proxy
            # class instead of the abstract class to handle return values for this type.
            if inspect.isabstract(klass):
                klass = klass.__jsii_proxy_class__()

            # Create our instance, bypassing __init__ by directly calling __new__, and
            # then assign our reference to __jsii_ref__
            inst = klass.__new__(klass)
            inst.__jsii_ref__ = ref

            if ref.interfaces is not None:
                return InterfaceDynamicProxy(
                    [inst] + self.build_interface_proxies_for_ref(ref)
                )
            else:
                return inst

        # Legacy code path - Kernel invariant ought to guarantee that class_fqn can't be Struct (they're interfaces)
        elif class_fqn in _data_types:
            # Data types have been serialized by-reference (see aws/jsii#400).
            # We retrieve all of its properties right now and then construct a value
            # object from it. This will be slow :(.

            # Ugly delayed import here because I can't solve the cyclic
            # package dependency right now :(.
            from ._runtime import python_jsii_mapping

            data_type = _data_types[class_fqn]
            remote_struct = _FakeReference(ref)

            python_props = {
                python_name: kernel.get(remote_struct, jsii_name)
                for python_name, jsii_name in (
                    python_jsii_mapping(data_type) or {}
                ).items()
            }

            return data_type(**python_props)
        elif class_fqn in _enums:
            return _enums[class_fqn]
        elif class_fqn == "Object":
            # If any one interface is a struct, all of them are guaranteed to be (Kernel invariant)
            if ref.interfaces is not None and any(
                fqn in _data_types for fqn in ref.interfaces
            ):
                # Ugly delayed import here because I can't solve the cyclic
                # package dependency right now :(.
                from ._runtime import python_jsii_mapping

                structs = [_data_types[fqn] for fqn in ref.interfaces]
                remote_struct = _FakeReference(ref)

                if len(structs) == 1:
                    struct = structs[0]
                else:
                    struct = new_combined_struct(structs)

                return struct(
                    **{
                        python_name: kernel.get(remote_struct, jsii_name)
                        for python_name, jsii_name in (
                            python_jsii_mapping(struct) or {}
                        ).items()
                    }
                )
            else:
                return InterfaceDynamicProxy(self.build_interface_proxies_for_ref(ref))
        else:
            raise ValueError(f"Unknown type: {class_fqn}")

    def resolve_id(self, id: str) -> Any:
        return self._refs[id]

    def build_interface_proxies_for_ref(self, ref: ObjRef) -> List[Any]:
        ifaces = [_interfaces[fqn] for fqn in ref.interfaces or []]
        classes = [iface.__jsii_proxy_class__() for iface in ifaces]

        # If there's no classes, use an Opaque reference to make sure the
        # __jsii_ref__ property is visible through the InterfaceDynamicProxy.
        if len(classes) == 0:
            return [Opaque(ref)]

        insts = [klass.__new__(klass) for klass in classes]
        for inst in insts:
            inst.__jsii_ref__ = ref
        return insts


class Opaque:
    def __init__(self, ref: ObjRef) -> None:
        # Set the __jsii_type__ property on the class if it's not there already
        if getattr(self.__class__, "__jsii_type__", None) is None:
            setattr(self.__class__, "__jsii_type__", "Object")

        # Track the jsii reference
        self.__jsii_ref__ = ref


class InterfaceDynamicProxy(object):
    def __init__(self, delegates):
        self._delegates = delegates

    def __getattr__(self, name):
        for delegate in self._delegates:
            if hasattr(delegate, name):
                return getattr(delegate, name)
        type_info = "+".join([str(delegate.__class__) for delegate in self._delegates])
        raise AttributeError(f"'%s' object has no attribute '%s'" % (type_info, name))

    def __setattr__(self, name, value):
        if name == "_delegates":
            return super().__setattr__(name, value)
        for delegate in self._delegates:
            if hasattr(delegate, name):
                return setattr(delegate, name, value)
        type_info = "+".join([str(delegate.__class__) for delegate in self._delegates])
        raise AttributeError(f"'%s' object has no attribute '%s'" % (type_info, name))


def new_combined_struct(structs: Iterable[Type]) -> Type:
    label = " + ".join(struct.__name__ for struct in structs)

    def __init__(self, **kwargs):
        self._values = kwargs

    def __eq__(self, rhs: Any) -> bool:
        return isinstance(rhs, self.__class__) and rhs._values == self._values

    def __ne__(self, rhs: Any) -> bool:
        return not (rhs == self)

    def __repr__(self) -> str:
        return f"<{label}>({', '.join(k + '=' + repr(v) for k, v in self._values.items())})"

    return type(
        label,
        (*structs,),
        {
            "__init__": __init__,
            "__eq__": __eq__,
            "__ne__": __ne__,
            "__repr__": __repr__,
        },
    )


_refs = _ReferenceMap(_types)


register_reference = _refs.register
resolve_reference = _refs.resolve
resolve_id = _refs.resolve_id



import pytest
import re
from typing import Optional

from scope.jsii_calc_lib.custom_submodule_name import NestingClass
from scope.jsii_calc_lib import Number
import jsii_calc
import jsii


class TestRuntimeTypeChecking:
    """
    These tests verify that runtime type checking performs the necessary validations and produces error messages that
    are indicative of the error. There are #type:ignore annotations scattered everywhere as these tests are obviously
    attempting to demonstrate what happens when invalid calls are being made.
    """

    def test_constructor(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument initial_value must be one of (int, float, NoneType); got str instead"
            ),
        ):
            jsii_calc.Calculator(initial_value="nope")  # type:ignore

    def test_constructor_decorated(self):
        """
        This test verifies that runtime type checking is not broken when a function is wrapped with a decorator, as was
        the case with the original implementation (due to a dynamic access to type_hints for the method via the type).
        """

        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument maximum_value must be one of (int, float, NoneType); got str instead"
            ),
        ):
            orig_init = jsii_calc.Calculator.__init__

            # For toy, swap initial_value and maximum_values here
            def decorated(
                self,
                *,
                initial_value: Optional[jsii.Number] = None,
                maximum_value: Optional[jsii.Number] = None,
            ):
                orig_init(
                    self, initial_value=maximum_value, maximum_value=initial_value
                )

            jsii_calc.Calculator.__init__ = decorated
            try:
                jsii_calc.Calculator(initial_value="nope")  # type:ignore
            finally:
                jsii_calc.Calculator.__init__ = orig_init

    def test_struct(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument foo must be jsii_calc.StringEnum; got int instead"
            ),
        ):
            jsii_calc.StructWithEnum(foo=1337)  # type:ignore

    def test_method_arg(self):
        subject = jsii_calc.Calculator()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument value must be one of (int, float); got str instead"
            ),
        ):
            subject.mul("Not a Number")  # type:ignore

    def test_method_kwarg(self):
        subject = jsii_calc.DocumentedClass()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument name must be one of (str, NoneType); got int instead"
            ),
        ):
            subject.greet(name=1337)  # type:ignore

    def test_method_vararg(self):
        subject = jsii_calc.StructPassing()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument inputs[0] must be jsii_calc.TopLevelStruct; got int instead"
            ),
        ):
            subject.how_many_var_args_did_i_pass(1337, 42)  # type:ignore

    def test_setter_to_enum(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument value must be jsii_calc.AllTypesEnum; got int instead"
            ),
        ):
            subject.enum_property = 1337  # type:ignore

    def test_setter_to_primitive(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape("type of argument value must be str; got int instead"),
        ):
            subject.string_property = 1337  # type:ignore

    def test_setter_to_map(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument value must be collections.abc.Mapping; got jsii_calc.StructWithEnum instead"
            ),
        ):
            subject.map_property = jsii_calc.StructWithEnum(  # type:ignore
                foo=jsii_calc.StringEnum.A
            )

    def test_setter_to_list(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape("type of argument value must be a list; got int instead"),
        ):
            subject.array_property = 1337  # type:ignore

    def test_setter_to_list_with_invalid_value(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape("type of argument value[0] must be str; got int instead"),
        ):
            subject.array_property = [1337]  # type:ignore

    def test_setter_to_union(self):
        subject = jsii_calc.AllTypes()
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument value must be one of (str, int, float, scope.jsii_calc_lib.Number, jsii_calc.Multiply); got jsii_calc.StringEnum instead"
            ),
        ):
            subject.union_property = jsii_calc.StringEnum.B  # type:ignore

    def test_nested_struct(self):
        # None of these should throw...
        NestingClass.NestedStruct(name="Queen B")

    def test_anonymous_object(self):
        struct = jsii_calc.StructUnionConsumer.provide_struct("A")
        assert jsii_calc.StructUnionConsumer.is_struct_a(struct)

        iface = jsii_calc.anonymous.UseOptions.provide("A")
        assert jsii_calc.anonymous.UseOptions.consume(iface) == "A"

    def test_nested_union(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument union_property[0] must be one of (Mapping[str, Union[jsii_calc.StructA, Dict[str, Any], jsii_calc.StructB]], Sequence[Union[jsii_calc.StructA, Dict[str, Any], jsii_calc.StructB]]); got float instead"
            ),
        ):
            jsii_calc.ClassWithNestedUnion([1337.42])  # type:ignore

    def test_variadic(self):
        with pytest.raises(
            TypeError,
            match=re.escape(
                "type of argument union[1] must be one of (jsii_calc.StructA, jsii_calc.StructB); got float instead"
            ),
        ):
            jsii_calc.VariadicTypeUnion(
                jsii_calc.StructA(required_string="present"), 1337.42  # type:ignore
            )

        jsii_calc.VariadicTypeUnion()

    def test_homonymous_forward_references(self):
        """Verifies homonymous forward references don't trip the type checker

        This has been an issue when stub functions were introduced to create a reliable source for type checking
        information, which was reported in https://github.com/aws/jsii/issues/3818.
        """
        # This uses a ForwardRef["Homonymous"] that should resolve to jsii_calc.homonymous_forward_references.foo.Homonymous
        jsii_calc.homonymous_forward_references.foo.Consumer.consume(
            homonymous={"string_property": "Check!"}
        )
        # This uses a ForwardRef["Homonymous"] that should resolve to jsii_calc.homonymous_forward_references.bar.Homonymous
        jsii_calc.homonymous_forward_references.bar.Consumer.consume(
            homonymous={"numeric_property": 1337}
        )

    def test_shadowed_namespaces_are_not_a_problem(self):
        """Verifies that a parameter shadowing a namespace does not cause errors

        This has caused https://github.com/aws/aws-cdk/issues/22975.
        """

        subject = jsii_calc.ParamShadowsScope()
        num = Number(1337)
        # The parameter is named "scope" which shadows the "scope" module...
        assert num == subject.use_scope(num)

    def test_shadowed_builtins_are_not_a_problem(self):
        """Verifies that a parameter shadowing a built-in name does not cause errors"""

        jsii_calc.ParamShadowsBuiltins(
            "${not a Python type (builtins)}",
            "${not a Python type (str)}",
            string_property="Most definitely a string",
            boolean_property=True,
            struct_property={
                "required_string": "Present!",
            },
        )



import atexit
import base64
import datetime
import contextlib
import enum
import json
import os
import os.path
import pathlib
import platform
import subprocess
import sys
import tempfile
import threading

from typing import TYPE_CHECKING, Type, Union, Mapping, IO, Any, AnyStr, Optional

import attr
import cattr  # type: ignore
import dateutil.parser

import jsii._embedded.jsii

from ...__meta__ import __jsii_runtime_version__
from ..._compat import importlib_resources
from ..._utils import memoized_property
from .base import BaseProvider
from ..types import (
    ObjRef,
    EnumRef,
    Override,
    KernelRequest,
    KernelResponse,
    LoadRequest,
    LoadResponse,
    CreateRequest,
    CreateResponse,
    DeleteRequest,
    DeleteResponse,
    GetRequest,
    GetResponse,
    InvokeRequest,
    InvokeResponse,
    GetScriptCommandRequest,
    GetScriptCommandResponse,
    InvokeScriptRequest,
    InvokeScriptResponse,
    SetRequest,
    SetResponse,
    StaticGetRequest,
    StaticInvokeRequest,
    StaticSetRequest,
    BeginRequest,
    BeginResponse,
    EndRequest,
    EndResponse,
    CallbacksRequest,
    CallbacksResponse,
    CompleteRequest,
    CompleteResponse,
    StatsRequest,
    StatsResponse,
    Callback,
    CompleteRequest,
    CompleteResponse,
)
from ...errors import ErrorType, JSIIError, JavaScriptError


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _HelloResponse:
    hello: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _OkayResponse:
    # We could technically mark this as KernelResponse, because we know that
    # it is going to be one of those. However, we can't disambiguate the different
    # types because some of them have the same keys as each other, so the only way
    # to know what type the result is expected to be, is to know what method is
    # being called. Thus we'll expect Any here, and structure this value separately.
    ok: Any


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _ErrorResponse:
    error: str
    stack: str
    name: str


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _CallbackResponse:
    callback: Callback


@attr.s(auto_attribs=True, frozen=True, slots=True)
class _CompleteRequest:
    complete: CompleteRequest


_ProcessResponse = Union[_OkayResponse, _ErrorResponse, _CallbackResponse]


def _with_api_key(api_name, asdict):
    def unstructurer(value):
        unstructured = asdict(value)
        unstructured["api"] = api_name

        return unstructured

    return unstructurer


def _with_reference(data, type_):
    if not isinstance(data, type_):
        return type_(ref=data.ref)
    return data


def _unstructure_ref(value):
    return {"$jsii.byref": value.ref}


def _unstructure_enum(member):
    return {"$jsii.enum": f"{member.__class__.__jsii_type__}/{member.value}"}


def ohook(d):
    if d.keys() == {"$jsii.byref"} or d.keys() == {"$jsii.byref", "$jsii.interfaces"}:
        return ObjRef(ref=d["$jsii.byref"], interfaces=d.get("$jsii.interfaces"))
    if d.keys() == {"$jsii.date"}:
        return dateutil.parser.isoparse(d["$jsii.date"])
    if d.keys() == {"$jsii.enum"}:
        ref, member = d["$jsii.enum"].rsplit("/", 1)
        return EnumRef(ref=ObjRef(ref=ref + "@"), member=member)
    if d.keys() == {"$jsii.map"}:
        return d["$jsii.map"]
    return d


def jdefault(obj):
    if hasattr(obj, "__jsii_ref__"):
        return _unstructure_ref(obj.__jsii_ref__)
    if isinstance(obj, datetime.datetime) and obj.tzinfo is not None:
        return {"$jsii.date": obj.isoformat()}
    elif isinstance(obj, datetime.datetime):
        raise TypeError("Naive datetimes are not supported, please add a timzone.")
    raise TypeError("Don't know how to convert object to JSON: %r" % obj)


class _NodeProcess:
    def __init__(self):
        self._serializer = cattr.Converter()
        self._serializer.register_unstructure_hook(enum.Enum, _unstructure_enum)
        self._serializer.register_unstructure_hook(
            LoadRequest,
            _with_api_key("load", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            GetScriptCommandRequest,
            _with_api_key(
                "getBinScriptCommand", self._serializer.unstructure_attrs_asdict
            ),
        )
        self._serializer.register_unstructure_hook(
            InvokeScriptRequest,
            _with_api_key("invokeBinScript", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            CreateRequest,
            _with_api_key("create", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            DeleteRequest,
            _with_api_key("del", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            GetRequest, _with_api_key("get", self._serializer.unstructure_attrs_asdict)
        )
        self._serializer.register_unstructure_hook(
            StaticGetRequest,
            _with_api_key("sget", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            SetRequest, _with_api_key("set", self._serializer.unstructure_attrs_asdict)
        )
        self._serializer.register_unstructure_hook(
            StaticSetRequest,
            _with_api_key("sset", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            InvokeRequest,
            _with_api_key("invoke", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            StaticInvokeRequest,
            _with_api_key("sinvoke", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            BeginRequest,
            _with_api_key("begin", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            EndRequest, _with_api_key("end", self._serializer.unstructure_attrs_asdict)
        )
        self._serializer.register_unstructure_hook(
            CallbacksRequest,
            _with_api_key("callbacks", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            CompleteRequest,
            _with_api_key("complete", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            StatsRequest,
            _with_api_key("stats", self._serializer.unstructure_attrs_asdict),
        )
        self._serializer.register_unstructure_hook(
            Override, self._serializer.unstructure_attrs_asdict
        )
        self._serializer.register_unstructure_hook(ObjRef, _unstructure_ref)
        self._serializer.register_structure_hook(ObjRef, _with_reference)

        self._ctx_stack = contextlib.ExitStack()

    def __del__(self):
        self.stop()

    def _jsii_runtime(self) -> str:
        tmpdir = self._ctx_stack.enter_context(tempfile.TemporaryDirectory())
        resources = {
            resname: os.path.join(tmpdir, filename.replace("/", os.sep))
            for resname, filename in jsii._embedded.jsii.EMBEDDED_FILES.items()
        }

        for resname, filename in resources.items():
            pathlib.Path(os.path.dirname(filename)).mkdir(exist_ok=True)
            with open(filename, "wb") as fp:
                fp.write(
                    importlib_resources.files(jsii._embedded.jsii)
                    .joinpath(resname)
                    .read_bytes()
                )

        # Return our first path, which should be the path for jsii-runtime.js
        return resources[jsii._embedded.jsii.ENTRYPOINT]

    def _next_message(self) -> Mapping[Any, Any]:
        assert self._process.stdout is not None
        return json.loads(self._process.stdout.readline(), object_hook=ohook)

    def start(self):
        environ = os.environ.copy()
        environ["JSII_AGENT"] = f"Python/{platform.python_version()}"

        jsii_node = environ.get("JSII_NODE", "node")
        jsii_runtime = environ.get("JSII_RUNTIME", self._jsii_runtime())

        self._process = subprocess.Popen(
            [
                jsii_node,
                "--max-old-space-size=4069",
                jsii_runtime,
            ],
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            env=environ,
        )

        self.sink_thread = threading.Thread(
            name="process.stderr_sink",
            target=stderr_sink,
            # Trailing comma here is important (this is a 1-value tuple, not a value between parentheses)
            args=(self._process.stderr,),
            # Thread is a daemon so it does not hold the VM from shutting down
            daemon=True,
        )
        self.sink_thread.start()

        # Clean this process up at exit, so it terminates "gracefully"
        atexit.register(self.stop)

        self.handshake()

    def stop(self) -> None:
        # This process is closing already, un-registering the hook to not fire twice
        atexit.unregister(self.stop)

        assert self._process.stdin is not None
        if not self._process.stdin.closed:
            self._process.stdin.write(b'{"exit":0}\n')
            # Close the process' STDIN, singaling we are done with it
            self._process.stdin.close()

        try:
            self._process.wait(timeout=5)
        except subprocess.TimeoutExpired:
            self._process.terminate()

        if self.sink_thread.is_alive():
            self.sink_thread.join(timeout=5)

        self._ctx_stack.close()

    def handshake(self) -> None:
        # Get the version of the runtime that we're using.
        resp: _HelloResponse = self._serializer.structure(
            self._next_message(), _HelloResponse
        )

        # TODO: Replace with proper error.
        assert (
            resp.hello == f"@jsii/runtime@{__jsii_runtime_version__}"
            # Transparently allow development versions of the runtime to be used.
            or resp.hello == f"@jsii/runtime@0.0.0"
        ), f"Invalid JSII Runtime Version: {resp.hello!r}"

    def send(
        self, request: KernelRequest, response_type: Type[KernelResponse]
    ) -> KernelResponse:
        req_dict = self._serializer.unstructure(request)
        data = json.dumps(req_dict, default=jdefault).encode("utf8")

        # Send our data, ensure that it is framed with a trailing \n
        assert self._process.stdin is not None
        self._process.stdin.write(b"%b\n" % (data,))
        self._process.stdin.flush()

        resp: _ProcessResponse = self._serializer.structure(
            self._next_message(), _ProcessResponse
        )

        if isinstance(resp, _OkayResponse):
            return self._serializer.structure(resp.ok, response_type)
        elif isinstance(resp, _CallbackResponse):
            return resp.callback
        else:
            if resp.name == ErrorType.JSII_FAULT.value:
                raise JSIIError(resp.error) from JavaScriptError(resp.stack)
            raise RuntimeError(resp.error) from JavaScriptError(resp.stack)


class ProcessProvider(BaseProvider):
    @memoized_property
    def _process(self) -> _NodeProcess:
        process = _NodeProcess()
        process.start()

        return process

    def load(self, request: LoadRequest) -> LoadResponse:
        return self._process.send(request, LoadResponse)

    def getScriptCommand(
        self, request: GetScriptCommandRequest
    ) -> GetScriptCommandResponse:
        return self._process.send(request, GetScriptCommandResponse)

    def invokeBinScript(self, request: InvokeScriptRequest) -> InvokeScriptResponse:
        return self._process.send(request, InvokeScriptResponse)

    def create(self, request: CreateRequest) -> CreateResponse:
        return self._process.send(request, CreateResponse)

    def get(self, request: GetRequest) -> GetResponse:
        return self._process.send(request, GetResponse)

    def set(self, request: SetRequest) -> SetResponse:
        return self._process.send(request, SetResponse)

    def sget(self, request: StaticGetRequest) -> GetResponse:
        return self._process.send(request, GetResponse)

    def sset(self, request: StaticSetRequest) -> SetResponse:
        return self._process.send(request, SetResponse)

    def invoke(self, request: InvokeRequest) -> Union[InvokeResponse, Callback]:
        return self._process.send(request, InvokeResponse)

    def sinvoke(self, request: StaticInvokeRequest) -> InvokeResponse:
        return self._process.send(request, InvokeResponse)

    def delete(self, request: DeleteRequest) -> DeleteResponse:
        return self._process.send(request, DeleteResponse)

    def begin(self, request: BeginRequest) -> BeginResponse:
        return self._process.send(request, BeginResponse)

    def end(self, request: EndRequest) -> EndResponse:
        return self._process.send(request, EndResponse)

    def callbacks(self, request: CallbacksRequest) -> CallbacksResponse:
        return self._process.send(request, CallbacksResponse)

    def complete(self, request: CompleteRequest) -> CompleteResponse:
        return self._process.send(request, CompleteResponse)

    def sync_complete(
        self, request: CompleteRequest, response_type: Type[KernelResponse]
    ) -> Union[InvokeResponse, GetResponse]:
        resp = self._process.send(_CompleteRequest(complete=request), response_type)
        return resp

    def stats(self, request: Optional[StatsRequest] = None) -> StatsResponse:
        if request is None:
            request = StatsRequest()
        return self._process.send(request, StatsResponse)


def stderr_sink(reader: IO[AnyStr]) -> None:
    # An empty string is used to signal EOF...
    for line in iter(reader.readline, b""):
        if line == b"":
            break
        try:
            console = json.loads(line)
            if console.get("stderr") is not None:
                sys.stderr.buffer.write(base64.b64decode(console["stderr"]))
            if console.get("stdout") is not None:
                sys.stdout.buffer.write(base64.b64decode(console["stdout"]))
        except:
            print(line, file=sys.stderr)
